/*
   USB-HID Volume COntroller for V-USB

   Copyright (C) 2013 morecat_lab
   based on V-USB

*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include "usbdrv.h"
#include <avr/eeprom.h>
#include <avr/wdt.h>

/* for two LED */
#define REVERSE_CLICK_LED

/* for Play / NextTrac / PrevTrac hack */
#undef PLAY_CONTROL

static uchar reportBuffer[2] = {1, 0} ;
/*
 reportBuffer[0] :  Report_ID (always 1)
 reportBuffer[1] :  Report Data ( Volume Up, Volume Down, Mute)
*/
static uchar idleRate;           /* in 4 ms units */

/* LED is on PB5 */
#define LED_ON   { PORTB |=   (1 << 5); }
#define LED_OFF  { PORTB &= ~ (1 << 5); }
#define LED_FLIP { PORTB = ((PORTB & (1 << 5)) == 0) ?  (PORTB |= (1 << 5)) : (PORTB &= ~(1 << 5)); }

 /* USB report descriptor */
PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {
  0x05, 0x0c,                    // USAGE_PAGE (Consumer Devices)
  0x09, 0x01,                    // USAGE (Consumer Control)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x01,			 // REPORT_ID = 1
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x09, 0xe9,                    //   USAGE (Volume Up)
  0x09, 0xea,                    //   USAGE (Volume Down)
  0x75, 0x01,                    //   REPORT_SIZE (1)
  0x95, 0x02,                    //   REPORT_COUNT (2)
  0x81, 0x02,			 //   INPUT (Data, Var Abs)
  0x09, 0xe2,                    //   USAGE (Mute)
  0x09, 0xb6,                    //   USAGE (Scan Previous Track)
  0x09, 0xb5,                    //   USAGE (Scan Next Track)
  0x09, 0xb7,                    //   USAGE (Stop)
  0x09, 0xb1,                    //   USAGE (Pause)
  0x09, 0xb0,                    //   USAGE (Play)
  0x95, 0x06,                    //   REPORT_COUNT (6)
  0x81, 0x06,                    //   INPUT (Data, Var, Real)
  0xc0                           // END_COLLECTION
};


/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void) {
  uchar step = 128;
  uchar trialValue = 0, optimumValue;
  int   x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

  /* do a binary search: */
  do{
    OSCCAL = trialValue + step;
    x = usbMeasureFrameLength();    /* proportional to current real frequency */
    if(x < targetValue)             /* frequency still too low */
      trialValue += step;
    step >>= 1;
  }while(step > 0);
  /* We have a precision of +/- 1 for optimum OSCCAL here */
  /* now do a neighborhood search for optimum value */
  optimumValue = trialValue;
  optimumDev = x; /* this is certainly far away from optimum */
  for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
    x = usbMeasureFrameLength() - targetValue;
    if(x < 0)
      x = -x;
    if(x < optimumDev){
      optimumDev = x;
      optimumValue = OSCCAL;
    }
  }
  OSCCAL = optimumValue;
}

/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

void    usbEventResetReady(void) {
  calibrateOscillator();
  eeprom_write_byte(0, OSCCAL);   /* store the calibrated value in EEPROM */
}



/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8]) {
  usbRequest_t    *rq = (void *)data;

  usbMsgPtr = reportBuffer;
  if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
    if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
      /* we only have one report type, so don't look at wValue */
      return sizeof(reportBuffer);
    }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
      usbMsgPtr = &idleRate;
      return 1;
    }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
      idleRate = rq->wValue.bytes[1];
    }
  }else{
    /* no vendor specific requests implemented */
  }
  return 0;
}

void hardwareInit(void) {
  uchar    i, j;

  /* activate pull-ups except on USB lines */
  USB_CFG_IOPORT   = (uchar)~((1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT));
  /* all pins input except USB (-> USB reset) */
#ifdef USB_CFG_PULLUP_IOPORT    /* use usbDeviceConnect()/usbDeviceDisconnect() if available */
  USBDDR    = 0;    /* we do RESET by deactivating pullup */
  usbDeviceDisconnect();
#else
  USBDDR    = (1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT);
#endif

  j = 0;
  while(--j){          /* USB Reset by device only required on Watchdog Reset */
    i = 0;
    while(--i);      /* delay >10ms for USB reset */
  }
#ifdef USB_CFG_PULLUP_IOPORT
  usbDeviceConnect();
#else
  USBDDR    = 0;      /*  remove USB reset condition */
#endif
}

void portInit() {
  DDRB  = 0b00100000;		/* PB5 = OUT */
#ifdef REVERSE_CLICK_LED
  PORTB = 0b00011001;		/* LED OFF , PULL-UP PB4,PB3,PB0 */
#else
  PORTB = 0b00111001;		/* LED ON , PULL-UP PB4,PB3,PB0 */
#endif
}

/* HID Report format */
#define VolumeUp   (1 << 0)
#define VolumeDown (1 << 1)
#define Mute       (1 << 2)
#define PrevTrac   (1 << 3)
#define NextTrac   (1 << 4)
#define Stop       (1 << 5)
#define Pause      (1 << 6)
#define Play       (1 << 7)

/* blink interval */
#define BLINK_COUNT_MAX (2000)
#define CHECK_INTERVAL (25)

/* チャッタリング防止用にキーを読み込む待ち時間 */
#define BUTTON_COUNT_TH (8)

uint8_t createHIDReport() {

  uint8_t status = 0;
  static uint8_t dataRead = 0;
#ifdef PLAY_CONTROL
  static uint8_t state_pause = 0;
#else
  static uint8_t state_mute = 0;
#endif
  static int8_t button_cnt = BUTTON_COUNT_TH;;
  static uint16_t blink_cnt = 0;
  int8_t direction;
  static const int dir[] = { 0,1,-1,0,0,0,0,0,0,0,0,0,0,0,0,0 }; /* 回転方向テーブル(１クリック１アップダウン） */

  dataRead = (dataRead << 2) + ((PINB >> 3) & 0b00000011);  /* 前回値と今回値でインデックスとする */
  direction = dir[dataRead & 15];

  if (direction == 1) {
#ifdef PLAY_CONTROL
    status = NextTrac;
    state_pause = 1;
#else
    status = VolumeUp;
    state_mute = 0;
#endif    
#ifdef REVERSE_CLICK_LED
    LED_OFF;
#else
    LED_ON;
#endif
  } else if (direction == -1) {
#ifdef PLAY_CONTROL
    status = PrevTrac;
    state_pause = 1;
#else
    status = VolumeDown;
    state_mute = 0;
#endif
#ifdef REVERSE_CLICK_LED
    LED_OFF;
#else
    LED_ON;
#endif
  }
  if (status != 0)
    return status;
    
  
  if ((PINB & 0x1) == 0 ) { /* PB0 for Mute */
    if (button_cnt >= 0) {
      button_cnt--;
      if (button_cnt < 0) {
#ifdef PLAY_CONTROL
	status |= Play;
#else
	status |= Mute;
#endif
#ifdef PLAY_CONTROL
	if (state_pause == 0) {
	  state_pause = 1;
#else
	if (state_mute == 0) {
	  state_mute = 1;	/* MUTE ON */
#endif
#ifdef REVERSE_CLICK_LED
	  LED_ON;
#else
	  LED_OFF;
#endif
	} else {
#ifdef PLAY_CONTROL
	  state_pause = 0;
#else
	  state_mute = 0;	/* MUTE OFF */
#endif
#ifdef REVERSE_CLICK_LED
	  LED_OFF;
#else
	  LED_ON;
#endif
	}
      }
    }
  } else {
    button_cnt = BUTTON_COUNT_TH;
  }
  
#ifdef PLAY_CONTROL
  if (state_pause == 1) {
#else
  if (state_mute == 1) {
#endif
    if ((--blink_cnt) == 0) {
      LED_FLIP;
      blink_cnt = BLINK_COUNT_MAX;
    }
  } else {
    blink_cnt = BLINK_COUNT_MAX;
  }
  
  return status;
}


int main(void) {
  uchar calibrationValue;
  uint8_t state = 0;
  uint8_t preStatus = 0;
  uchar i = CHECK_INTERVAL;
  
  calibrationValue = eeprom_read_byte(0); /* calibration value from last time */
  if(calibrationValue != 0xff){
    OSCCAL = calibrationValue;
  }

  hardwareInit();
  usbInit();
  portInit();

  sei();                  
  wdt_enable(WDTO_8S);
	
  for(;;){                /* main event loop */
    if (usbConfiguration != 0)
      wdt_reset(); //reset wdt only when USB connection is established

    usbPoll();          
    
    if (i-- == 0) {
      state = createHIDReport();
      i = CHECK_INTERVAL;
    }

    if (state != preStatus) {
      if (usbInterruptIsReady()){
	reportBuffer[1] = state;
	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
	preStatus = state;
      }
    }
  }
  return 0;
}

/* EOF */
